"""
    Fastberry Command-Line-Interface
"""

import click

# Import Commands Here
from .app_manager import start_app
from .graphql import graphql
from .run import run
from .sql_alembic import cli

# Add Commands Here
cli.add_command(run)
cli.add_command(graphql)
cli.add_command(start_app)



"""
    SQLAlchemy + Alembic
"""

import click

try:
    from alembic import command
    from alembic.config import Config

    ALEMBIC_CONFIG = Config("alembic.ini")

except ImportError:
    ALEMBIC_CONFIG = None


@click.group()
def cli():
    """Click (CLI) Group"""


@cli.command()
@click.option("-m", "--message", help="Migration message.", type=str, default=None)
def db_migrate(message):
    """Database Make-Migrations & Migrate."""
    # Make-Migrations
    command.revision(ALEMBIC_CONFIG, message=message, autogenerate=True)
    # Migrate
    command.upgrade(ALEMBIC_CONFIG, "head")


@cli.command()
@click.option("-r", "--revision", help="Revision ID.", type=str)
def db_upgrade(revision):
    """Database Migrate (Upgrade)."""
    command.upgrade(ALEMBIC_CONFIG, revision)


@cli.command()
@click.option("-r", "--revision", help="Revision ID.", type=str)
def db_downgrade(revision):
    """Database Migrate (Downgrade)."""
    command.downgrade(ALEMBIC_CONFIG, revision)


@cli.command()
def db_history():
    """Database Migrations History."""
    command.history(ALEMBIC_CONFIG)



"""
SQL-Alchemy ORM to Custom-Types Mapper
"""
import functools
import types
import typing

from .scalars import SQLALCHEMY_ACTIVE, SQLALCHEMY_FIELDS, Manager, sql_many_to_many


def related_field(type_config, key, related_class_name):
    """
    Get Related Fields
    """
    # is_many_to_many = table_config.many_to_many.get(key)
    table_config = type_config.table_config
    related_model = Manager.find_model(related_class_name)
    related_table = related_model.__meta__.table_name
    if not related_model.__meta__.sql:
        # Mongo Database
        return SQLALCHEMY_FIELDS.Column(
            SQLALCHEMY_FIELDS.String,
            index=table_config.index(key),
            unique=table_config.unique(key),
            primary_key=table_config.primary_key(key),
            nullable=not table_config.required(key),
        )
    # SQL Database
    return SQLALCHEMY_FIELDS.Column(
        SQLALCHEMY_FIELDS.Integer,
        SQLALCHEMY_FIELDS.ForeignKey(f"{related_table}._id"),
        index=table_config.index(key),
        unique=table_config.unique(key),
        primary_key=table_config.primary_key(key),
        nullable=not table_config.required(key),
    )


# sqlalchemy_base
def related_many_to_many(sqlalchemy_base, type_config, related_class_name):
    """
    Get Related Many-To-Many Fields
    """
    # is_many_to_many = table_config.many_to_many.get(key)
    related_model = Manager.find_model(related_class_name)
    related_table = related_model.__meta__.table_name
    return sql_many_to_many(sqlalchemy_base, type_config.table_name, related_table)


def create_sqlalchemy_model(base, type_config):
    """
    Create SQLAlchemy Model
    """
    if SQLALCHEMY_ACTIVE:
        sqlalchemy_base = base
        sqlalchemy_class_name = f"{type_config.name}SQLAlchemy"
        table_config = type_config.table_config
        # Table Setup
        sqlalchemy_class_setup = {
            "__tablename__": type_config.table_name,
            "__table_args__": [],
            "_id": SQLALCHEMY_FIELDS.Column(
                SQLALCHEMY_FIELDS.Integer, primary_key=True
            ),
        }
        sqlalchemy_setup_lazy_load = {}
        sqlalchemy_many_to_many_lazy_load = {}
        sqlalchemy_many_to_many = {}
        one_or_more_sqlalchemy_lazy_load = False
        # Get Fields
        for key, setup in type_config.custom_annotations.items():
            sqlalchemy_lazy_load = False
            if not setup.types:
                column_name = f"{key}_id"
                if isinstance(setup.real, typing.ForwardRef):
                    related_class_name = setup.real.__forward_arg__
                elif isinstance(setup.real, str):
                    related_class_name = setup.real
                related_model = Manager.find_model(related_class_name)
                if not related_model:
                    sqlalchemy_lazy_load = True
                    one_or_more_sqlalchemy_lazy_load = True
                # Many-To-Many Relationship or Regular Relationship?
                if not key in table_config.many_to_many:
                    # Lazy Load or Load Now?
                    if sqlalchemy_lazy_load:
                        sqlalchemy_setup_lazy_load[column_name] = functools.partial(
                            related_field, type_config, key, related_class_name
                        )
                    else:
                        sqlalchemy_class_setup[column_name] = related_field(
                            type_config, key, related_class_name
                        )
                else:
                    if sqlalchemy_lazy_load:
                        do_many_to_many = functools.partial(
                            related_many_to_many,
                            sqlalchemy_base[0],
                            type_config,
                            related_class_name,
                        )
                        sqlalchemy_many_to_many_lazy_load[key] = do_many_to_many
                    else:
                        do_many_to_many = related_many_to_many(
                            sqlalchemy_base[0],
                            type_config,
                            related_class_name,
                        )
                        sqlalchemy_many_to_many[key] = do_many_to_many

            else:
                sqlalchemy_class_setup[key] = SQLALCHEMY_FIELDS.Column(
                    setup.types.sql,
                    index=table_config.index(key),
                    unique=table_config.unique(key),
                    primary_key=table_config.primary_key(key),
                    nullable=not table_config.required(key),
                )

        for config in table_config.unique_together:
            unique_together = SQLALCHEMY_FIELDS.UniqueTogether(*config)
            sqlalchemy_class_setup["__table_args__"].append(unique_together)

        sqlalchemy_class_setup["__table_args__"] = tuple(
            sqlalchemy_class_setup["__table_args__"]
        )

        class LoadLater:
            """Load Objects (aka: SQLAlchemy ORM)"""

            def __init__(self):
                """Init Load-Later"""
                self.lazy = True
                self.objects = None
                self._relationships = None

                if not one_or_more_sqlalchemy_lazy_load:
                    self.lazy = False
                    # Objects
                    self.objects = type(
                        sqlalchemy_class_name, sqlalchemy_base, sqlalchemy_class_setup
                    ).__table__
                    # Relationship
                    self._relationships = types.SimpleNamespace(
                        **sqlalchemy_many_to_many
                    )
                else:
                    relationship_setup = sqlalchemy_many_to_many

                    def load_table(cls):
                        """When Ready Load Tables"""
                        for key, lazy_load in sqlalchemy_setup_lazy_load.items():
                            if callable(lazy_load):
                                sqlalchemy_class_setup[key] = lazy_load()
                            else:
                                sqlalchemy_class_setup[key] = lazy_load

                        # Objects
                        cls.objects = type(
                            sqlalchemy_class_name,
                            sqlalchemy_base,
                            sqlalchemy_class_setup,
                        ).__table__
                        for key, lazy_load in sqlalchemy_many_to_many_lazy_load.items():
                            if callable(lazy_load):
                                relationship_setup[key] = lazy_load()
                            else:
                                relationship_setup[key] = lazy_load
                        for key, loader in sqlalchemy_many_to_many.items():
                            relationship_setup[key] = loader

                        # Relationship
                        cls._relationships = types.SimpleNamespace(**relationship_setup)

                    # Objects
                    self.objects = load_table

        return LoadLater()
    return None



"""
    * ID (Encoder || Decoder)
"""

import base64
import hashlib


def encode(text):
    """[summary]
    Base64 Encoder.

    Args:
        text (str): String to be encoded.

    Returns:
        str: base64.urlsafe_b64encode
    """
    fill_up = hashlib.blake2s(f"{text}".encode("utf-8"), digest_size=8).hexdigest()
    text = f"{ text }::{ fill_up }"
    return base64.urlsafe_b64encode(text.encode("utf-8")).decode("utf-8")


def decode(text):
    """[summary]
    Base64 Decoder.

    Args:
        text (str): String to be decoded.

    Returns:
        str: base64.urlsafe_b64decode
    """
    code = base64.urlsafe_b64decode(text.encode("utf-8")).decode("utf-8")
    text = code.split("::")
    return text[0]


class ID:
    """[summary]
    ID (Encoder || Decoder)
    """

    @staticmethod
    def encode(keys):
        """Encode-Wrapper"""
        try:
            if not isinstance(keys, list):
                return encode(keys)
            return [encode(key) for key in keys]
        except Exception as error:
            raise Exception("Invalid ID") from error

    @staticmethod
    def decode(keys):
        """Decode-Wrapper"""
        try:
            if not isinstance(keys, list):
                return decode(keys)
            return [decode(key) for key in keys]
        except Exception as error:
            raise Exception("Invalid ID") from error


def sql_id_decode(unique_id) -> int | None:
    """Decoder for SQL"""
    try:
        return_value = int(ID.decode(unique_id))
    except Exception:
        return_value = None
    return return_value



"""Get App(s) Methods

Declared as <Query> or <Mutation>.
Each method is grouped by one of the two category.

methods:
    process_strawberry_crud: Collect App Methods.
"""

from ...utils.objects import get_attr, get_fields
from .gql import is_crud


def process_strawberry_crud(app_module):
    """Strawberry C.R.U.D"""

    gql_schema = {
        "Query": [],
        "Mutation": [],
        "Operations": {"Query": [], "Mutation": []},
    }
    for crud_name in get_fields(app_module):
        crud_module = getattr(app_module, crud_name)
        # CRUD
        if is_crud(crud_module):
            for field in [
                "Query",
                "Mutation",
            ]:
                op_class = get_attr(crud_module, field)
                op_fields = get_fields(op_class)
                for op in op_fields:
                    op_method = get_attr(op_class, op)
                    setattr(op_class, op, staticmethod(op_method))
                if op_class:
                    setattr(op_class, "keys", op_fields)
            # Init class and add all (Queries & Mutations)
            crud = crud_module()
            gql_query = get_attr(crud, "Query")
            gql_mutation = get_attr(crud, "Mutation")
            gql_names = get_attr(crud, "Operations")
            # Operations-Names
            gql_schema["Operations"]["Query"].extend(gql_names["Query"])
            gql_schema["Operations"]["Mutation"].extend(gql_names["Mutation"])
            # Items or None
            if gql_query:
                gql_schema["Query"].append(gql_query)
            if gql_mutation:
                gql_schema["Mutation"].append(gql_mutation)
    return gql_schema



from types import SimpleNamespace

from .ids import ID


def to_obj(items, sql: bool = False):
    """Convert Database-Based-Object into Python-SimpleNamespace-Object
    Args:
        items (list or dict): Database <Rows> or <Row>
        sql (bool, optional): IS-SQL Database? Defaults to False.
    """

    def row_handler(row: dict):
        # SQL-Alchemy
        row2dict = lambda row: {key: getattr(row, key) for key in row.keys()}
        if sql:
            row = row2dict(row)
        # Common
        unique_id = str(row["_id"])
        row["id"] = ID.encode(unique_id)
        if row:
            return SimpleNamespace(**row)
        return None

    def item_handler(rows):
        if isinstance(rows, list):
            # IF-List
            return [row_handler(i) for i in rows]
        # IF-Dict
        return row_handler(rows)

    if items:
        return item_handler(items)
    elif isinstance(items, list) and len(items) == 0:
        return []
    return None
